//
// Calculate optimal triangular prism decomposition (2 params optimzation)
//
#include <cstdio>
#include "R3Graph.h"
#include "tetrahed.h"
using namespace R3Graph;

int main() {
    double dz = sqrt(3.) / 2.;
    double dx = 1.;
    double dh = 0.0005;
    double h = 0.;      // Optimization parameter

    double minMeanAspect = 1e+30;
    double minMeanH = 0.;
    double minMeanDy = 0.;
    double meanAspect[3];

    double minMaxAspect = 1e+30;
    double minMaxH = 0.;
    double minMaxDy = 0.;
    double maxAspect[3];

    double dy = 0.75;    // Optimization parameter
    while (dy <= 1.25) {
        //
        //        p4
        //  p5
        //              p2
        //        p3
        //  p0
        //              p1
        //
        R3Point p1(dx, 0., 0.);
        R3Point p2(dx, dy, 0.);

        h = 0.;
        bool meanChanged = false;
        bool maxChanged = false;
        while (h <= dy / 2.) {
            R3Point p0(0., h, 0.);
            R3Point p3(dx/2., (p0.y + p2.y)/2., dz);
            R3Point p4(p3.x, p3.y + dy, dz);
            R3Point p5(0., p0.y + dy, 0.);

            Tetrahedron t0(p0, p1, p2, p3);
            Tetrahedron t1(p0, p2, p5, p3);
            Tetrahedron t2(p3, p4, p5, p2);

            double asp0 = t0.aspect();
            double asp1 = t1.aspect();
            double asp2 = t2.aspect();

            double meanAsp = (asp0 + asp1 + asp2) / 3.;
            double maxAsp = asp0;
            if (asp1 > maxAsp)
                maxAsp = asp1;
            if (asp2 > maxAsp)
                maxAsp = asp2;

            if (meanAsp < minMeanAspect) {
                minMeanAspect = meanAsp;
                minMeanH = h;
                minMeanDy = dy;
                meanAspect[0] = asp0;
                meanAspect[1] = asp1;
                meanAspect[2] = asp2;
                /*...
                printf(
                    "h=%lf dy=%lf asp0=%lf asp1=%lf asp2=%lf  mean asp=%lf\n",
                    h, dy, asp0, asp1, asp2, meanAsp
                );
                ...*/
                meanChanged = true;
            }

            if (maxAsp < minMaxAspect) {
                minMaxAspect = maxAsp;
                minMaxH = h;
                minMaxDy = dy;
                maxAspect[0] = asp0;
                maxAspect[1] = asp1;
                maxAspect[2] = asp2;
                /*...
                printf(
                    "h=%lf dy=%lf asp0=%lf asp1=%lf asp2=%lf  max asp=%lf\n",
                    h, dy, asp0, asp1, asp2, maxAsp
                );
                ...*/
                maxChanged = true;
            }

            h += dh;
        }

        if (meanChanged) {
            printf(
                "dy=%lf h=%lf asp0=%lf asp1=%lf asp2=%lf  mean asp=%lf\n",
                minMeanDy, minMeanH, meanAspect[0], meanAspect[1], meanAspect[2], minMeanAspect
            );
        }
        if (maxChanged) {
            printf(
                "dy=%lf h=%lf asp0=%lf asp1=%lf asp2=%lf  max asp=%lf\n",
                minMaxDy, minMaxH, maxAspect[0], maxAspect[1], maxAspect[2], minMaxAspect
            );
        }

        dy += dh;
    }

    printf("\n======================\n");

    printf(
        "dy=%lf h=%lf asp0=%lf asp1=%lf asp2=%lf  min. mean asp=%lf\n",
        minMeanDy, minMeanH, meanAspect[0], meanAspect[1], meanAspect[2], minMeanAspect
    );

    printf(
        "dy=%lf h=%lf asp0=%lf asp1=%lf asp2=%lf  min. max asp=%lf\n",
        minMaxDy, minMaxH, maxAspect[0], maxAspect[1], maxAspect[2], minMaxAspect
    );

    // Optimal tri-prism
    printf("\n======================\n");
    //... dy = minMeanDy;
    //... h = minMeanH;
    // Guessed:
    dy = 3. * sqrt(2.) / 4.;
    h = dy / 3.;

    R3Point p1(dx, 0., 0.);
    R3Point p2(dx, dy, 0.);
    R3Point p0(0., h, 0.);
    R3Point p3(dx/2., (p0.y + p2.y)/2., dz);
    R3Point p4(p3.x, p3.y + dy, dz);
    R3Point p5(0., p0.y + dy, 0.);

    Tetrahedron t0(p0, p1, p2, p3);
    Tetrahedron t1(p0, p2, p5, p3);
    Tetrahedron t2(p3, p4, p5, p2);

    double asp0 = t0.aspect();
    double asp1 = t1.aspect();
    double asp2 = t2.aspect();

    double meanAsp = (asp0 + asp1 + asp2) / 3.;
    printf("Optimal tri-prism decomposition:\n");
    printf(
        "h=%lf asp0=%lf asp1=%lf asp2=%lf  mean asp=%lf\n",
        h, asp0, asp1, asp2, meanAsp
    );

    printf("Tetrahedron 0:\n");
    printf("  Vertices:\n");
    printf("    (%lf, %lf, %lf)\n", t0.vertex[0].x, t0.vertex[0].y, t0.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t0.vertex[1].x, t0.vertex[1].y, t0.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t0.vertex[2].x, t0.vertex[2].y, t0.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t0.vertex[3].x, t0.vertex[3].y, t0.vertex[3].z);
    printf("  Edges:\n");
    printf("    %lf, %lf, %lf\n    %lf, %lf, %lf\n",
        t0.vertex[0].distance(t0.vertex[1]),
        t0.vertex[0].distance(t0.vertex[2]),
        t0.vertex[0].distance(t0.vertex[3]),
        t0.vertex[1].distance(t0.vertex[2]),
        t0.vertex[1].distance(t0.vertex[3]),
        t0.vertex[2].distance(t0.vertex[3])
    );

    printf("Tetrahedron 1:\n");
    printf("  Vertices:\n");
    printf("    (%lf, %lf, %lf)\n", t1.vertex[0].x, t1.vertex[0].y, t1.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t1.vertex[1].x, t1.vertex[1].y, t1.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t1.vertex[2].x, t1.vertex[2].y, t1.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t1.vertex[3].x, t1.vertex[3].y, t1.vertex[3].z);
    printf("  Edges:\n");
    printf("    %lf, %lf, %lf\n    %lf, %lf, %lf\n",
        t1.vertex[0].distance(t1.vertex[1]),
        t1.vertex[0].distance(t1.vertex[2]),
        t1.vertex[0].distance(t1.vertex[3]),
        t1.vertex[1].distance(t1.vertex[2]),
        t1.vertex[1].distance(t1.vertex[3]),
        t1.vertex[2].distance(t1.vertex[3])
    );

    printf("Tetrahedron 2:\n");
    printf("  Vertices:\n");
    printf("    (%lf, %lf, %lf)\n", t2.vertex[0].x, t2.vertex[0].y, t2.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t2.vertex[1].x, t2.vertex[1].y, t2.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t2.vertex[2].x, t2.vertex[2].y, t2.vertex[0].z);
    printf("    (%lf, %lf, %lf)\n", t2.vertex[3].x, t2.vertex[3].y, t2.vertex[3].z);
    printf("  Edges:\n");
    printf("    %lf, %lf, %lf\n    %lf, %lf, %lf\n",
        t2.vertex[0].distance(t2.vertex[1]),
        t2.vertex[0].distance(t2.vertex[2]),
        t2.vertex[0].distance(t2.vertex[3]),
        t2.vertex[1].distance(t2.vertex[2]),
        t2.vertex[1].distance(t2.vertex[3]),
        t2.vertex[2].distance(t2.vertex[3])
    );

    return 0;
}
